import com.sun.star.lang.*;
import com.sun.star.sheet.*;
import com.sun.star.table.*;
import com.sun.star.beans.XPropertySet;
import com.sun.star.uno.UnoRuntime;
import com.sun.star.uno.Any;
import com.sun.star.uno.AnyConverter;
import com.sun.star.uno.Type;
import com.sun.star.container.XIndexAccess;
import com.sun.star.container.XEnumerationAccess;
import com.sun.star.container.XEnumeration;

import java.util.*;
import java.util.regex.*;

import jp.ac.kobe_u.cs.cream.*;

public class CreamSolver {
    private static _CreamSolver solver = null;
    private static String creamCellRegex =
	"^\\s*=(\\s*jp\\.ac\\.kobe_u\\.cs\\.cream\\.uno\\.CreamAddins\\.\\w*\\s*\\(.*\\)\\s*)$";

    public static XEnumeration cellRangeEnumeration(XCellRange xCellRange) {
	XCellRangesQuery xRangesQuery = (XCellRangesQuery)UnoRuntime.
	    queryInterface(XCellRangesQuery.class, xCellRange);
	XEnumerationAccess xEA = xRangesQuery.
	    queryContentCells((short)CellFlags.FORMULA).getCells();
	return xEA.createEnumeration();
    }

    private static List getCreamCellsFromCellRange(XCellRange xCellRange) {
	List list = new ArrayList();
	XEnumeration xEnum = cellRangeEnumeration(xCellRange);
	Pattern pattern = Pattern.compile(creamCellRegex);
	try {
	    while (xEnum.hasMoreElements()) {
		Object aCell = xEnum.nextElement();
		XCell xCell = (XCell)UnoRuntime.
		    queryInterface(XCell.class, aCell);
		String formula = xCell.getFormula();
		Matcher matcher = pattern.matcher(formula);
		if (matcher.matches()) {
		    list.add(xCell);
		}
	    }
	} catch (com.sun.star.container.NoSuchElementException e) {
	} catch (com.sun.star.lang.WrappedTargetException e) {
	}
	return list;
    }

    private static XCellRange getCellRangeFromSheet(Object sheet) {
	XCellRange xCellRange = null;
	try {
	    XSpreadsheetDocument xSpreadsheetDocument = (XSpreadsheetDocument)UnoRuntime.
		queryInterface(XSpreadsheetDocument.class, sheet);
	    XIndexAccess xIndexAccess = (XIndexAccess)UnoRuntime.
		queryInterface(XIndexAccess.class, xSpreadsheetDocument.getSheets());
	    XSpreadsheet xSpreadsheet = (XSpreadsheet)UnoRuntime.
		queryInterface(XSpreadsheet.class, xIndexAccess.getByIndex(0));
	    xCellRange = (XCellRange)UnoRuntime.
		queryInterface(XCellRange.class, xSpreadsheet);
	} catch (com.sun.star.lang.IndexOutOfBoundsException e) {
	} catch (com.sun.star.lang.WrappedTargetException e) {
	}
	return xCellRange;
    }

    public static boolean CStart(Object sheet) {
	if (solver != null) {
	    solver.stop();
	}
	XCellRange xCellRange = getCellRangeFromSheet(sheet);
	solver = new _CreamSolver(xCellRange);
	List list = getCreamCellsFromCellRange(xCellRange);
	solver.addCreamCells(list);
	solver.start();
	solver.waitNext();
	Solution solution = solver.getSolution();
	if (solution == null) {
	    solver.stop();
	    solver = null;
	    return false;
	}
	solver.setSolution(solution);
	return true;
    }

    public static boolean CNext() {
	if (solver == null) {
	    return false;
	}
	solver.resume();
	solver.waitNext();
	Solution solution = solver.getSolution();
	if (solution == null) {
	    solver.stop();
	    solver = null;
	    return false;
	}
	solver.setSolution(solution);
	return true;
    }

    public static int CVAR(XPropertySet xOptions, Object domain, int value) {
	if (solver == null) {
	    return 0;
	}
	return value;
    }

    public static Object CINT(XPropertySet xOptions, int inf, int sup) {
	return "CINT(" + inf + ";" + sup + ")";
    }

    public static Object C(XPropertySet xOptions, Object constraint) {
	/*
	return toBoolean(constraint) ? "TRUE" : "FALSE";
	*/
	return "C()";
    }

    public static Object CNOTEQUALS(XPropertySet xOptions, Object[] cells) {
	/*
	boolean notEquals = true;
	try {
	    for (int i = 0; notEquals && i < cells.length; i++) {
		double x = AnyConverter.toDouble(cells[i]);
		for (int j = i + 1; notEquals && j < cells.length; j++) {
		    if (x == AnyConverter.toDouble(cells[j])) {
			notEquals = false;
		    }
		}
	    }
	} catch (com.sun.star.lang.IllegalArgumentException e) {
	    notEquals = false;
	}
	return notEquals ? "TRUE" : "FALSE";
	*/
	return "CNOTEQUALS()";
    }

    public static Object CSOLVE(XPropertySet xOptions, Object[] constraints) {
	return "CSOLVE()";
    }

    private static boolean toBoolean(Object uno) {
	boolean b = true;
	try {
	    if (uno instanceof String) {
		b = uno.equals("TRUE");
	    } else if (AnyConverter.isDouble(uno)) {
		b = (AnyConverter.toDouble(uno) != 0.0);
	    }
	} catch (com.sun.star.lang.IllegalArgumentException e) {
	}
	return b;
    }


}


class _CreamSolver {
    private static long timeout = 30000;
    private static String variableFunctionName =
	"jp.ac.kobe_u.cs.cream.uno.CreamAddins.CVAR".toUpperCase();
    private static String intDomainFunctionName =
	"jp.ac.kobe_u.cs.cream.uno.CreamAddins.CINT".toUpperCase();
    private static String constraintFunctionName =
	"jp.ac.kobe_u.cs.cream.uno.CreamAddins.C".toUpperCase();
    private static String notEqualsFunctionName =
	"jp.ac.kobe_u.cs.cream.uno.CreamAddins.CNOTEQUALS".toUpperCase();
    private static String solveFunctionName =
	"jp.ac.kobe_u.cs.cream.uno.CreamAddins.CSOLVE".toUpperCase();
    private XCellRange xCellRange;
    private Hashtable variables;
    private Network network;
    private Solver solver = null;

    class CellVariable {
	public XCell xCell;
	public Variable variable;

	public CellVariable(XCell xCell, Variable variable) {
	    this.xCell = xCell;
	    this.variable = variable;
	}
    }

    public _CreamSolver(XCellRange xCellRange) {
	this.xCellRange = xCellRange;
    }

    private XCell getCell(CalcParser.Expression x) {
	XCell xCell = null;
	if (x.isSymbol()) {
	    try {
		xCell = xCellRange.getCellRangeByName(x.op.str).getCellByPosition(0, 0);
	    } catch (com.sun.star.lang.IndexOutOfBoundsException e) {
	    }
	}
	return xCell;
    }

    private String getCellKey(XCell xCell) {
	String key = null;
	if (xCell != null) {
	    XCellAddressable xCellAddr = (XCellAddressable)UnoRuntime.
		queryInterface(XCellAddressable.class, xCell);
	    CellAddress aAddress = xCellAddr.getCellAddress();
	    key = aAddress.Column + " " + aAddress.Row;
	}
	return key;
    }

    private int getCellValue(CalcParser.Expression x) {
	int value = 0;
	if (x.isInteger()) {
	    value = Integer.parseInt(x.op.str);
	} else if (x.isSymbol()) {
	    XCell xCell = getCell(x);
	    String key = getCellKey(xCell);
	    if (key != null && variables.get(key) == null) {
		value = (int)xCell.getValue();
	    }
	}
	return value;
    }

    private Variable getCellVariable(XCell xCell) {
	Variable v = null;
	String key = getCellKey(xCell);
	if (key != null && variables.get(key) != null) {
	    v = ((CellVariable)variables.get(key)).variable;
	//} else if (! xCell.getType().equals(CellContentType.EMPTY)) {
	} else {
	    v = new IntVariable(network, (int)xCell.getValue());
	}
	return v;
    }

    private Variable sum(CalcParser.Expression[] args) {
	if (args == null) {
	    return new IntVariable(network, 0);
	}
	IntVariable sum = null;
	for (int i = 0; i < args.length; i++) {
	    if (args[i] != null && args[i].isSymbol()) {
		XCellRange xr = xCellRange.getCellRangeByName(args[i].op.str);
		XEnumeration xEnum = CreamSolver.cellRangeEnumeration(xr);
		try {
		    while (xEnum.hasMoreElements()) {
			Object aCell = xEnum.nextElement();
			XCell xCell = (XCell)UnoRuntime.
			    queryInterface(XCell.class, aCell);
			IntVariable v = (IntVariable)getCellVariable(xCell);
			if (v != null) {
			    if (sum == null) {
				sum = v;
			    } else {
				sum = sum.add(v);
			    }
			}
		    }
		} catch (com.sun.star.container.NoSuchElementException e) {
		} catch (com.sun.star.lang.WrappedTargetException e) {
		}
	    }
	}
	return sum;
    }

    private Variable getTerm(CalcParser.Expression x) {
	Variable v = null;
	if (x.isInteger()) {
	    v = new IntVariable(network, Integer.parseInt(x.op.str));
	} else if (x.isSymbol()) {
	    v = getCellVariable(getCell(x));
	} else if (x.isFunctor(intDomainFunctionName, 2)) {
	    int inf = getCellValue(x.args[0]);
	    int sup = getCellValue(x.args[1]);
	    v = new IntVariable(network, inf, sup);
	} else if (x.isOperator("+", 2)) {
	    Variable[] vs = getTerms(x.args);
	    v = new IntVariable(network);
	    new IntArith(network, IntArith.ADD, v, vs[0], vs[1]);
	} else if (x.isOperator("-", 2)) {
	    Variable[] vs = getTerms(x.args);
	    v = new IntVariable(network);
	    new IntArith(network, IntArith.SUBTRACT, v, vs[0], vs[1]);
	} else if (x.isOperator("*", 2)) {
	    Variable[] vs = getTerms(x.args);
	    v = new IntVariable(network, IntDomain.FULL);
	    new IntArith(network, IntArith.MULTIPLY, v, vs[0], vs[1]);
	} else if (x.isFunctor("SUM")) {
	    v = sum(x.args);
	}
	return v;
    }

    private Variable[] getTerms(CalcParser.Expression[] xs) {
	List list = new ArrayList();
	for (int i = 0; i < xs.length; i++) {
	    Variable v = getTerm(xs[i]);
	    if (v != null) {
		list.add(v);
	    }
	}
	Variable[] vs = new Variable[1];
	vs = (Variable[])list.toArray(vs);
	return vs;
    }

    private Constraint notEquals(CalcParser.Expression x) {
	Constraint c = null;
	if (x.isSymbol()) {
	    List list = new ArrayList();
	    XCellRange xr = xCellRange.getCellRangeByName(x.op.str);
	    XEnumeration xEnum = CreamSolver.cellRangeEnumeration(xr);
	    try {
		while (xEnum.hasMoreElements()) {
		    Object aCell = xEnum.nextElement();
		    XCell xCell = (XCell)UnoRuntime.
			queryInterface(XCell.class, aCell);
		    Variable v = getCellVariable(xCell);
		    if (v != null) {
			list.add(v);
		    }
		}
	    } catch (com.sun.star.container.NoSuchElementException e) {
	    } catch (com.sun.star.lang.WrappedTargetException e) {
	    }
	    Variable[] vs = new Variable[1];
	    vs = (Variable[])list.toArray(vs);
	    c = new NotEquals(network, vs);
	}
	return c;
    }

    private void solveFunction(CalcParser.Expression[] args) {
	for (int i = 0; i < args.length; i++) {
	    CalcParser.Expression x = args[i];
	    List list = new ArrayList();
	    XCellRange xr = xCellRange.getCellRangeByName(x.op.str);
	    XEnumeration xEnum = CreamSolver.cellRangeEnumeration(xr);
	    try {
		while (xEnum.hasMoreElements()) {
		    Object aCell = xEnum.nextElement();
		    XCell xCell = (XCell)UnoRuntime.
			queryInterface(XCell.class, aCell);
		    CalcParser parser = new CalcParser(xCell.getFormula());
		    CalcParser.Expression y = parser.parse();
		    addConstraint(y);
		}
	    } catch (com.sun.star.container.NoSuchElementException e) {
	    } catch (com.sun.star.lang.WrappedTargetException e) {
	    }
	}
    }

    private void addConstraint(CalcParser.Expression x) {
	Constraint c = null;
	if (x.isOperator("=", 2)) {
	    Variable[] vs = getTerms(x.args);
	    if (vs.length == 2) {
		c = new Equals(network, vs);
	    }
	} else if (x.isOperator("<>", 2)) {
	    Variable[] vs = getTerms(x.args);
	    if (vs.length == 2) {
		c = new NotEquals(network, vs);
	    }
	} else if (x.isOperator("<", 2)) {
	    Variable[] vs = getTerms(x.args);
	    if (vs.length == 2) {
		c = new IntComparison(network, IntComparison.LT, vs[0], vs[1]);
	    }
	} else if (x.isOperator("<=", 2)) {
	    Variable[] vs = getTerms(x.args);
	    if (vs.length == 2) {
		c = new IntComparison(network, IntComparison.LE, vs[0], vs[1]);
	    }
	} else if (x.isOperator(">", 2)) {
	    Variable[] vs = getTerms(x.args);
	    if (vs.length == 2) {
		c = new IntComparison(network, IntComparison.GT, vs[0], vs[1]);
	    }
	} else if (x.isOperator(">=", 2)) {
	    Variable[] vs = getTerms(x.args);
	    if (vs.length == 2) {
		c = new IntComparison(network, IntComparison.GE, vs[0], vs[1]);
	    }
	} else if (x.isFunctor(notEqualsFunctionName, 1)) {
	    c = notEquals(x.args[0]);
	}
    }

    private void addConstraintCell(XCell xCell) {
	CalcParser parser = new CalcParser(xCell.getFormula());
	try {
	    CalcParser.Expression x = parser.parse();
	    if (x.isFunctor(solveFunctionName)) {
		solveFunction(x.args);
	    } else if (x.isFunctor(constraintFunctionName, 1)) {
		addConstraint(x.args[0]);
		/*
		x = x.args[0];
		Constraint c = null;
		if (x.isOperator("=", 2)) {
		    Variable[] vs = getTerms(x.args);
		    if (vs.length == 2) {
			c = new Equals(network, vs);
		    }
		} else if (x.isOperator("<>", 2)) {
		    Variable[] vs = getTerms(x.args);
		    if (vs.length == 2) {
			c = new NotEquals(network, vs);
		    }
		} else if (x.isOperator("<", 2)) {
		    Variable[] vs = getTerms(x.args);
		    if (vs.length == 2) {
			c = new IntComparison(network, IntComparison.LT, vs[0], vs[1]);
		    }
		} else if (x.isOperator("<=", 2)) {
		    Variable[] vs = getTerms(x.args);
		    if (vs.length == 2) {
			c = new IntComparison(network, IntComparison.LE, vs[0], vs[1]);
		    }
		} else if (x.isOperator(">", 2)) {
		    Variable[] vs = getTerms(x.args);
		    if (vs.length == 2) {
			c = new IntComparison(network, IntComparison.GT, vs[0], vs[1]);
		    }
		} else if (x.isOperator(">=", 2)) {
		    Variable[] vs = getTerms(x.args);
		    if (vs.length == 2) {
			c = new IntComparison(network, IntComparison.GE, vs[0], vs[1]);
		    }
		} else if (x.isFunctor(notEqualsFunctionName, 1)) {
		    c = notEquals(x.args[0]);
		}
		*/
	    }
	} catch (java.lang.IllegalArgumentException e) {
	}
    }

    private void addVariableCell(XCell xCell) {
	CalcParser parser = new CalcParser(xCell.getFormula());
	try {
	    CalcParser.Expression x = parser.parse();
	    if (x.isFunctor(variableFunctionName, 2)) {
		String key = getCellKey(xCell);
		if (variables.get(key) == null) {
		    Variable v = getTerm(x.args[0]);
		    if (v != null) {
			variables.put(key, new CellVariable(xCell, v));
		    }
		}
	    }
	} catch (java.lang.IllegalArgumentException e) {
	}
    }

    public void addCreamCells(List list) {
	network = new Network();
	variables = new Hashtable();
	Iterator cs = list.iterator();
	while (cs.hasNext()) {
	    XCell xCell = (XCell)cs.next();
	    addVariableCell(xCell);
	}
	cs = list.iterator();
	while (cs.hasNext()) {
	    XCell xCell = (XCell)cs.next();
	    addConstraintCell(xCell);
	}
	/*
        try {
            xCellRange.getCellByPosition(0, 0).setFormula(network.toString());
        } catch (com.sun.star.lang.IndexOutOfBoundsException e) {
        }
	*/
    }

    public void setSolution(Solution solution) {
	Pattern pattern = Pattern.compile(";[^;]*\\)\\s*$");
	Iterator cvs = variables.values().iterator();
	while (cvs.hasNext()) {
	    CellVariable cv = (CellVariable)cvs.next();
	    IntVariable v = (IntVariable)cv.variable;
	    int value = solution.getIntValue(v);
	    Matcher matcher = pattern.matcher(cv.xCell.getFormula());
	    String formula = matcher.replaceFirst(";" + value + ")");
	    cv.xCell.setFormula(formula);
	}
    }

    public void start() {
	if (solver != null) {
	    solver.stop();
	}
	solver = new DefaultSolver(network);
	solver.start(timeout);
    }

    public void waitNext() {
	if (solver != null) {
	    solver.waitNext(timeout);
	}
    }

    public void resume() {
	if (solver != null) {
	    solver.resume();
	}
    }

    public void stop() {
	if (solver != null) {
	    solver.stop();
	    solver = null;
	}
    }

    public Solution getSolution() {
	return solver.getSolution();
    }

}
